# -*- coding: utf-8 -*-

# FOGLAMP_BEGIN
# See: https://foglamp-foglamp-documentation.readthedocs-hosted.com
# FOGLAMP_END

""" Module for creating an webcam image recognition flow """

import os
import sys

import copy
from datetime import datetime

import logging
from foglamp.common import logger, iprpc
from foglamp.plugins.common import utils

from foglamp.common.plugin_helpers import HandleMap, PluginHandle

__author__ = "Douglas Orr"
__copyright__ = "Copyright (c) 2019 Dianomic Systems Inc."
__license__ = "Apache 2.0"
__version__ = "${VERSION}"


_module_dir = os.path.dirname(__file__)
_handle_map = HandleMap(__name__)

# SETTINGS
POLL_MODE = True  # -> poll(t) or async(f)   # XXX async not yet implemented
_plugin_mode = {True: "poll", False: "async"}[POLL_MODE]
_LOGGER = logger.setup(__name__, level=logging.DEBUG)


class MediaHandle(PluginHandle):
    RPC_SERVER_NAME='webcam_media_server'          # name of RPC server module
    RPC_CONFIG_MEMBERS=['fpm', 'camera_number', 'full_media_dir']
    
    def __init__(self, config, service_name):
        """[hold the state used to manage the process of capturing and analyzing images]

        Args:
            config ([type]): [The config from the plugin_init API.]
        """
        super().__init__(service_name)

        self.logger.setLevel(logging.DEBUG)

        # transfer parameters to handle members
        self.config_update(config)
        
        self.rpc = None
        self._setup(config)

        self.logger.info("DIR: {} ".format(self.full_media_dir))

    def config_update(self, update):
        """[Updated config to consider plugin init is called.]

        Args:
            update ([type]): [The config to consider.]

        Raises:
            ValueError: [if the directory mentioned does not exist and the plugin 
            is not called for the first time.]

        Returns:
            [boolean]: [False]
        """
        super().config_update(update)  # update the values in the handle

        # class-specific handle updates

        # root us in the FOGLAMP_DATA dir (xxx - canonicalize, avoid .. mischief)
        self.logger.info("data: {}  dir: {}".format(MediaHandle.FOGLAMP_DATA, self.media_dir))
        self.full_media_dir = os.path.join(MediaHandle.FOGLAMP_DATA, self.media_dir)

        _LOGGER.debug("the full directory is {}".format(self.full_media_dir))

        self.data_format = "IMG"

    def _setup(self, config):
        """[Either initializes an iprpc server client if camera is selcted or uses a directory
        of images if directory is selected from the media_type config value.]

        Args:
            config ([type]): [description]
        """
        if self.media_type == "camera":
            os.makedirs(self.full_media_dir, exist_ok=True)
            self.poll = self._cam_poll
            self.rpc_setup(config, _module_dir)
        else:
            if not os.path.exists(self.full_media_dir):
                raise ValueError("{}: does not exist".format(self.full_media_dir))
            # just taking images from a directory
            self.poll = self._dir_poll

            if self.rpc is not None:
                self.rpc.plugin_shutdown()

            self.rpc = None
            self.images = os.listdir(self.full_media_dir)

    def reconfigure(self, new_config):
        """[Handles reconfigure for the plugin side.]

        Args:
            new_config ([dict]): [The new configuration to be used.]
        """
        # assign handle values from new config
        _old_media_type = self.media_type
        
        self.config_update(new_config)

        if self.media_type != _old_media_type:
            # if we are going from dir -> camera or camera -> dir, we need back end changes
            # either get a new rpc server, or destroy the old rpc server, and create the new config
            self.shutdown()
                
            self._setup(new_config)
        elif self.media_type == "camera":
            # reconfiguring existing camera; all relevant changes happen on the back end
            self.rpc.plugin_init(self._rpc_config())
        else:
            # refresh the set of images we will "serve"
            self.images = os.listdir(self.full_media_dir)

    def shutdown(self):
        """[calls plugin_shutdown at the server side.]
        """
        if self.rpc is not None:
            self.rpc.plugin_shutdown()
            self.rpc = None

    def _cam_poll(self):
        """[Called when camera is selected from the media_type config type]

        Raises:
            ValueError: [If iprpc object was not initialized.]

        Returns:
            [type]: [The image names are converted into file names.]
        """
        if self.rpc is None:
            raise ValueError("internal handle does not exist")

        _frames = self.rpc.plugin_poll()

        if len(_frames) == 0:
            return None

        # create assets with the list camera frames  "dir": self.media_dir,

        _readings = [{
            'asset': self.asset_name,
            "timestamp": utils.local_timestamp(),
            'readings': {
                "datafmt": "IMG",
                "name": f,
                "timestamp": utils.local_timestamp(),
                "dir": self.media_dir
            }
        } for f in _frames]

        return _readings

    def _dir_poll(self):
        """[Called when directory is selected from the media_type config type]

        Returns:
            [dict]: [The image names are converted into file names.]
        """

        if len(self.images) == 0:
            if self.repeat_loop:
                # refresh the set of images we will "serve"
                self.images = os.listdir(self.full_media_dir)
            else:
                return None

        # create an asset with the next image  "dir": self.media_dir,
        _reading = {
            "asset": self.asset_name,
            "timestamp": utils.local_timestamp(),
            "readings": {
                "datafmt": self.data_format,
                "timestamp": utils.local_timestamp(),
                "name": self.images.pop(0),
                "dir": self.media_dir
            },
        }

        return _reading


MediaHandle.FOGLAMP_ROOT = os.getenv("FOGLAMP_ROOT", default="/usr/local/foglamp")
MediaHandle.FOGLAMP_DATA = os.path.join(MediaHandle.FOGLAMP_ROOT, "data")

# ### PLUGIN API ### #

_DEFAULT_CONFIG = {
    "plugin": {
        "description": "Webcam image recognition flow",
        "type": "string",
        "default": "webcam_media",
        "readonly": "true",
    },
    "assetName": {
        "description": "Name of Asset output",
        "type": "string",
        "default": "WebcamImages",
        "displayName": "Asset Name",
        "order": "1",
    },
    "mediaType": {
        "description": "Source media generated from",
        "type": "enumeration",
        "options": ["camera", "directory"],
        "default": "directory",
        "displayName": "Media type",
        "order": "2",
    },
    "mediaDir": {
        "description": "Directory where media will be stored",
        "type": "string",
        "default": "webcam",
        "displayName": "Media storage directory",
        "order": "3",
    },
    "dataFormat": {
        "description": "Format of files in 'dir'",
        "type": "enumeration",
        "options": ["IMG", "FRAME"],
        "default": "IMG",
        "displayName": "File data format",
        "order": "4",
        "validity": "mediaType == \"directory\"",
    },
    "repeatLoop": {
        "description": "Reload when you hit the end",
        "type": "boolean",
        "default": "false",
        "displayName": "Repeat loop",
        "order": "5",
        "validity": "mediaType == \"directory\"",
    },
    "cameraNumber": {
        "description": "Number of /dev/video camera device",
        "type": "integer",
        "default": "0",
        "displayName": "Camera number",
        "order": "6",
        "validity": "mediaType == \"camera\"",
    },
    "fpm": {
        "description": "frames per minute",
        "type": "float",
        "default": "10",
        "displayName": "Frames per minute processed",
        "order": "7",
        "validity": "mediaType == \"camera\"",
    },
}


def plugin_info():
    """Returns information about the plugin.
    Args:
    Returns:
        dict: plugin information
    Raises:
    """
    return {
        "name": "Webcam Image Flow",
        "version": "3.0.0",
        "mode": _plugin_mode,
        "type": "south",
        "interface": "1.0",
        "config": _DEFAULT_CONFIG,
    }


def plugin_init(config):
    """Initialise the plugin.
    Args:
        config: JSON configuration document for the South plugin configuration category
    Returns:
        handle: JSON object to be used in future calls to the plugin
    Raises:
    """
    _LOGGER.info("PLUGIN INIT, making a new media handle")
    _media_handle = MediaHandle(config, "webcam")
    _h = _handle_map.new_handle(_media_handle, config)
    return _h


def plugin_poll(h):
    """[The plugin poll mandatory method of FogLAMP plugin API.]

    Args:
        h ([dict]): [The incoming handle]

    Returns:
        [type]: [Returns reading after a poll]
    """
    _media_handle = _handle_map.get_handle(h)
    # vector off to the particular medium
    return _media_handle.poll()


def plugin_reconfigure(h, new_config):
    """Reconfigures the plugin

    Args:
        h: handle returned by the plugin initialisation call
        new_config: JSON object representing the new configuration category for the category
    Returns:
        new_handle: new handle to be used in the future calls
    """
    _media_handle = _handle_map.get_handle(h)

    _media_handle.reconfigure(new_config)  # may create/destroy rpc server

    h["config"].update(new_config)

    return h


def plugin_shutdown(h):
    """Shutdowns the plugin doing required cleanup, to be called prior to the South plugin service being shut down.

    Args:
        handle: handle returned by the plugin initialisation call
    Returns:
        plugin shutdown
    """
    _LOGGER.info("Web CAM media plugin shutdown")
    _media_handle = _handle_map.get_handle(h)
    _media_handle.shutdown()  # shutdown any RPC objects

    _handle_map.del_handle(h)  # remove our mapping


def plugin_register_ingest(h, callback, ingest_ref):
    """Required plugin interface component to communicate to South C server async mode

    Args:
        handle: handle returned by the plugin initialisation call
        callback: C opaque object required to passed back to C->ingest method
        ingest_ref: C opaque object required to passed back to C->ingest method
    """
    _LOGGER.info("Registering callback")
    _media_handle = _handle_map.get_handle(h)
    _media_handle.cb = (callback, ingest_ref)
